@@ -6,7 +6,7 @@
  * Copyright (C) 2002-2011
  *  Ludovic Rousseau <ludovic.rousseau@free.fr>
  *
- * $Id: pcscdaemon.c 6383 2012-07-01 16:11:52Z rousseau $
+ * $Id: pcscdaemon.c 6446 2012-08-24 09:07:18Z rousseau $
  */
 
 /**
@@ -406,6 +406,7 @@ int main(int argc, char **argv)
 	if (!setToForeground)
 	{
 		int pid;
+		int fd;
 
 		if (pipe(pipefd) == -1)
 		{
@@ -422,9 +423,17 @@ int main(int argc, char **argv)
 
 		/* like in daemon(3): redirect standard input, standard output
 		 * and standard error to /dev/null */
-		(void)close(0);
-		(void)close(1);
-		(void)close(2);
+		fd = open("/dev/null", O_RDWR);
+		if (fd != -1)
+		{
+			dup2(fd, STDIN_FILENO);
+			dup2(fd, STDOUT_FILENO);
+			dup2(fd, STDERR_FILENO);
+
+			/* do not close stdin, stdout or stderr */
+			if (fd > 2)
+				close(fd);
+		}
 
 		if (pid)
 		/* in the father */
@@ -588,8 +597,10 @@ int main(int argc, char **argv)
 	 * Set up the search for USB/PCMCIA devices
 	 */
 	rv = HPSearchHotPluggables();
+#ifndef USE_SERIAL
 	if (rv)
 		at_exit();
+#endif
 
 	rv = HPRegisterForHotplugEvents();
 	if (rv)
